import unicodedata

from math import log
from .util import tr_lower
from .vocabulary import Vocabulary
from .word_tokenizer import BertTokenizer


def word_shape(text):
    if len(text) >= 100:
        return "LONG"
    shape = []
    last = ""
    shape_char = ""
    seq = 0
    for char in text:
        if char.isalpha():
            if char.isupper():
                shape_char = "X"
            else:
                shape_char = "x"
        elif char.isdigit():
            shape_char = "d"
        else:
            shape_char = char
        if shape_char == last:
            seq += 1
        else:
            seq = 0
            last = shape_char
        if seq < 4:
            shape.append(shape_char)
    return "".join(shape)


class Token:
    vocabulary = None
    cache = {}

    @classmethod
    def _get_cache(cls, word):
        token = Token.cache.get(word, None)

        return token

    @classmethod
    def set_vocabulary(cls, tokenizer):
        cls.cache.clear()
        cls.vocabulary = Vocabulary.load(tokenizer.__name__)

    def __new__(cls, word, *args, **kwargs):
        if cls.vocabulary is None:
            cls.vocabulary = Vocabulary.load(BertTokenizer.__name__)

        cached = cls._get_cache(word)

        if cached:
            return cached

        token = super(Token, cls).__new__(cls)

        return token

    def __init__(self, word):
        if self in self.cache:
            return

        self.cache[word] = self

        self.word = word
        self.lower_ = tr_lower(word)
        self.is_punct = all(unicodedata.category(c).startswith("P") for c in word)
        self.is_digit = word.isdigit()
        self.shape = word_shape(word)

        self.entry = None

    def smooth_idf(self):
        return log(Token.vocabulary.document_count / (1 + self.df)) + 1

    @property
    def idf(self):
        return self.smooth_idf()

    @property
    def id(self):
        if self.entry is None:
            self.entry = Token.vocabulary[self.word]

        return self.entry.id

    @property
    def is_oov(self):
        return self.id == -1

    @property
    def df(self):
        if self.entry is None:
            self.entry = Token.vocabulary[self.word]

        return self.entry.df



from tqdm import tqdm
from collections import defaultdict
from itertools import islice
import click

from .. import Doc
from ..dataset.extended import load_extended_sents_corpus
from .util import tr_lower
from .vocabulary import Vocabulary


@click.command()
@click.option('--max-doc', help="Maximum documentation in extended corpus", type=int, default=-1)
@click.option('--min-df', help="Mininum document frequenct of a word to be included in", default=3)
@click.option('--word-tokenizer', type=click.Choice(['bert'], case_sensitive=False),
              help="Word tokenizer to be used in building vocabulary.", default='bert')
def build_vocabulary(max_doc, min_df, word_tokenizer):
    """Build vocabulary.
    """
    if max_doc > 0:
        corpus = islice(load_extended_sents_corpus(), max_doc)
    else:
        corpus = load_extended_sents_corpus()

    vocab = Vocabulary.factory(word_tokenizer)

    click.secho(click.style(f"...Frequency calculation over extended dataset", fg="blue"))

    for i, d in tqdm(enumerate(corpus), unit=" doc"):
        doc = Doc.from_sentences(d['sentences'])

        for sent in doc:
            for word in sent.tokens:
                vocab.add_word_to_doc(word, i)

    vocab.build(min_df)
    vocab.save()

    click.secho(click.style(f"Total documents {vocab.document_count}", fg="blue"))
    click.secho(click.style(f"Vocabulary size {len(vocab)} (words occured more than {min_df} documents)", fg="blue"))


if __name__ == '__main__':
    build_vocabulary()
